' SD Boot Test
' ------------
' by Andy Schenk (Ariba)



' ------ Board Configuration ----------
con
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

   do  = 0      '<-- sd card pins
   clk = 1
   di  = 2
   cs  = 3
' ------- end of configuration --------


'' The SD booter expects a shortcut in sector 0 (mostly MBR) to find the file.
'' This should work with any file system and also raw SD blocks (no file system).
'' The shortcut is 1 byte at $1D2 and and 2 longs at position $1D6 and $1DA:
'' $1D2: a signature byte: $27
'' $1D6: the startsector of the file to boot
'' $1DA: the number of contiguous sectors to load (8 is always safe for a min cluster size of 4kB)
''
'' This location is in fact the table data of partition 2 in a MBR partition table. So
'' you can also set up your SD cards with 2 partitions and have the bootcode in the second
'' partition with type $27. You should then use only menu item 5 and 6 of this test utility.
''
'' This version tries to find a file with the file signature at begin, if no shortcut exists
'' in sector 0. This allows to make an SD card bootable without any utility, just copy the boot-
'' file to a fresh formated SD card. You can even have two bootfiles, one that starts by searching
'' and another that starts when the shortcut is set.
'' For this test the file signature is: [P2BOOT]  (but only the first 4 characters are testet here)


DAT  'SD boot code (something like this may be in P2-ROM)  141 longs
                        org     0
entry
                        mov     blocksh,#0
                        or      outa,clkmask            'init pins
                        mov     dira,outmask            

                        neg     txdat,#1
                        mov     count,h200              '512x8 clocks for card reset
initSD                  call    #spiio
                        djnz    count,#initSD

                        mov     time,cnt                'start time for timeout (1 sec)
                        mov     count,#10               'try several times:
:loop                   mov     sdcmd,#0                ' go-idle command
                        mov     sdpar,#0
                        call    #spicmd
                        cmp     data,#1         wz      'until idle bit set
        if_z            jmp     #tstSD2
                        djnz    count,#:loop
                        
noBoot                  neg     data,#9                 '-> no responding SD card 
                        jmp     #fback

tstSD2                  mov     sdcmd,#8                'test if SD Type2 card
                        mov     sdpar,#$1AA
                        call    #spicmd
                        cmp     data,#1         wz
        if_nz           jmp     #normalSD               'no: -> normal SD (or MMC)
                        mov     bits,#32
                        call    #spibits                'yes: read supported voltage (no check)

                        mov     y,bit30                 'CMD41 with SDHC bit set
                        call    #cmd41                  'CMD41 = bring in operational mode
                        mov     sdcmd,#58               'read OCR (to check HC mode)
                        mov     sdpar,#0
                        call    #spicmd
                        mov     bits,#32
                        call    #spibits
                        test    data,bit30      wz      'in SDHC mode?
        if_z            mov     blocksh,#9              'no: SD address shift
                        jmp     #begin

normalSD                mov     y,#0                    'operational mode for SD Type1 card
                        call    #cmd41
                        mov     blocksh,#9              '0=SDHC 9=SD (address shift)
        
' check if pointer in sector 0 (FAT boot record or MBR)
begin                   mov     blocknr,#0
                        mov     ptra,hubbuf
                       add     ptra,#2 
                        call    #rdblock                'read sector 0
                        mov     ptra,hubbuf
                        add     ptra,#$1D4              'to partition 2 in MBR table 
                        rdbyte  data,ptra               'partition type $27 ?
                        cmp     data,#$27       wz
        if_ne           jmp     #search
                        add     ptra,#4
                        rdlong  blocknr,ptra            'get partition/file startsector
                        add     ptra,#4
                        rdlong  x,ptra          wz      'get number of contiguous blocks to boot
'        if_z            jmp     #search+6               'disable search if valid type but size=0
                        jmp     #found

' if no pointer in sector 0: search for sector with file signature at begin (maybe check a config cell to disable this)
search                  mov     y,#7
                        mov     x,#8                    'fixed size of 4kB (maybe #252 for 126kB)
:loop                   mov     ptra,hubbuf
                        add     blocknr,y               'alternating +7, +1
                        xor     y,#6
                        cmp     blocknr,maxblocks  wc   'search max 19999 sectors
        if_ae           neg     data,#10
        if_ae           jmp     #fback                  'no signature = err -10
                        call    #rdblock
                        rdlong  data,hubbuf
                        cmp     sigFile,data    wz      'file signature in sector?
        if_ne           jmp     #:loop                
'                        add     blocknr,#1              'boot begin is next block
        
found                   mov     ptra,hubbuf             'load sectors from file into hubram
:load                   call    #rdblock
                        add     blocknr,#1  
                        djnz    x,#:load
' the P2 code would boot cog0 here!
' we just return the sector number to Spin on success in this test code
                        mov     data,blocknr            'okay
fback                   sub     hubbuf,#4               'feedback to Spin
                        wrlong  data,hubbuf
halt                    jmp     #halt
        
'----- Subroutines ----
rdblock
                        mov     count,h200              'Read a block (512 nytes)
                        mov     time,cnt
                        mov     sdcmd,#17               'CMD17 = read block
                        mov     sdpar,blocknr
                        shl     sdpar,blocksh           'address shift for normal SD
                        call    #spicmd
                        call    #rdresp                 'wait for StartToken ($FE)
rbyte
                        call    #spiio                  'read byte
                        wrbyte  data,ptra               'and write to hubram
                        add     ptra,#1                 'ptra++
                        djnz    count,#rbyte
                        mov     bits,#16                'ignore CRC16 at end
                        call    #spibits
                        or      outa,csmask             'deassert CS
rdblock_ret             ret


cmd41                   mov     sdcmd,#55               'wait until card out of idle
                        mov     sdpar,#0                'and is in operational mode
                        call    #spicmd
                        mov     sdcmd,#41
                        mov     sdpar,y                 'handels param for SD and SDHC
                        call    #spicmd
                        tjnz    data,#cmd41
cmd41_ret               ret


' SPI routines
spibyte                 rol     txdat,#24                'send byte from txdat[7..0]
spiio                   mov     bits,#8                  'send byte from txdat[31..24]
spibits                 mov     data,#0                  'send 1..32 bits from txdat
:bit                    andn    outa,clkmask
                        rol     txdat,#1        wc       'shift txdat out (sd_di = Prop_do)
                        muxc    outa,dimask
                        or      outa,clkmask
                        test    domask,ina      wc       'shift data in (sd_do = Prop_di)
                        rcl     data,#1
                        djnz    bits,#:bit
spibyte_ret
spiio_ret
spibits_ret             ret


spicmd                  or      outa,csmask             'send an SD command (6 bytes)
                        nop                             'deassert CS before new cmd 
                        andn    outa,csmask             'begin a cmd frame (CS=0)
                        neg     txdat,#1
                        mov     bits,#32
                        call    #spibits                'send 32 High bits to finish prev cmd
                        mov     txdat,sdcmd
                        add     txdat,#$40              'then CMD byte
                        call    #spibyte
                        mov     txdat,sdpar             'then 32bits parameter
                        mov     bits,#32
                        call    #spibits
                        mov     txdat,#$95              'then CRC (only for CMD0 and CMD8)
                        cmp     sdcmd,#8        wz
        if_z            mov     txdat,#$87
                        call    #spibyte
rdresp
                        neg     txdat,#1                'wait for R1 response (status)
                        call    #spiio
                        mov     bits,cnt                'timeout?
                        sub     bits,time
                        cmp     bits,clockfreq  wc      '~1 second
        if_ae           neg     data,#13
        if_ae           jmp     #fback
                        cmp     data,#$FF       wz      'repeat until response
        if_z            jmp     #rdresp 
spicmd_ret
rdresp_ret              ret

clockfreq               long    80_000_000              'timeout 1 second at 80MHz
bit30                   long    1<<30

domask                  long    1<<do
csmask                  long    1<<cs
clkmask                 long    1<<clk
dimask                  long    1<<di
outmask                 long    1<<clk | 1<<di | 1<<cs

sigFile                 byte    "[P2B"
maxblocks               long    19999
hubbuf                  long    $7000           'use the hub memory at $7000 as temp buffer

h200                    long    512                     'already defined in P2-ROM boot code

sdcmd                   res     1
sdpar                   res     1
blocksh                 res     1
blocknr                 res     1
txdat                   res     1

'already defined in P2-ROM boot code
count                   res     1
bits                    res     1
time                    res     1
data                    res     1
x                       res     1
y                       res     1
'P2 special registers
ptra                    res     1



obj
' ---- Test and Utility code ----
' none of the following code is needed
' for booting, just to test it.

  dbg : "FullDuplexSerial"
  sd  : "fsrw26x"

var
 long  cog, startsect
 byte  sectbuf[512]

pub Main : i | key, err, n1,n2, tm
'
  dbg.start(31,30,0,115200)
  repeat
    dbg.str(string(16,1,"press a key",13))
    waitcnt(clkfreq+cnt)
  until dbg.rxcheck => 0
'  directBoot            'uncomment this line to start the booter cog direct without FSRW started first

  err := \sd.mount_explicit(do,clk,di,cs)
  if err < 0
    dbg.str(string("mount error "))
    dbg.dec(err)
    dbg.str(string(13,"FSRW can not mount this card. We"))
    dbg.str(string(13,"try it with the ROM boot code..",13))
    directBoot
  else
    dbg.str(string("SD mounted!",13))
    dbg.str(string("First File:",13))
    \sd.opendir
    if \sd.nextFile(@sectbuf) => 0
      dbg.str(@sectbuf)
      dbg.str(string(" at sector: "))
      \sd.popen(@sectbuf,"r")
      dbg.dec(sd.getStartSector)
      dbg.str(string(" size: "))
      dbg.dec(sd.get_filesize)
      \sd.pclose
    dbg.tx(13)

  repeat
    dbg.str(string(13,"Boot Test Menu:"))
    dbg.str(string(13," (1) Write a boot file to SD card"))
    dbg.str(string(13," (2) show used sectors of file"))
    dbg.str(string(13," (3) Write the pointer in Sector 0"))
    dbg.str(string(13," (4) Remove the pointer in Sector 0"))
    dbg.str(string(13," (5) Try to boot"))
    dbg.str(string(13," (6) show pointer data in Sector 0"))
    dbg.str(string(13," (7) Remove all boot data"))
  
    key := dbg.rx
    case key
      "1": err := \sd.popen(string("boot.txt"),"w")
           \sd.pwrite(@text,strsize(@text)+1)
           repeat i from 1 to 31
             \sd.pwrite($6800,512)        'fill in some data to make a 16kB file                   
           \sd.pclose
           dbg.str(string(13,"file written!",13))

      "2": err := \sd.popen(string("boot.txt"),"r")
           ifnot err
             dbg.str(string(13,"boot file sectors: "))
             repeat i from 0 to sd.get_filesize>>9
               if i & 7 == 0
                 dbg.tx(13)
               n1 := n2
               n2 := sd.datablock
               if n2-1 <> n1
                 dbg.tx("-")
                 dbg.tx(" ")
                 err++
               dbg.dec(n2)
               \sd.pread($7000,512)
               dbg.tx(" ")
             sd.pclose
             if err > 1
               dbg.str(string(13,"sectors are not contiguous",13))
             else  
               dbg.str(string(13,"sectors are contiguous",13))

      "3": err := \sd.popen(string("boot.txt"),"r")
           ifnot err
             startsect := sd.getStartSector
             sd.pclose
             dbg.str(string(13,"boot file at sector: "))
             dbg.dec(startsect)
             \sd.readsector(0,@sectbuf)    'read sector 0
             i := "y"
             if sectbuf[$1D2] <> 0 and sectbuf[$1D2] <> $FF
               dbg.str(string(13,"There is already a partition 2! overwrite it?(y/n?"))
               i := dbg.rx
             if i == "y"
               byte[@sectbuf+$1D2] := $27    'write boot partition type
               word[@sectbuf+$1D6] := startsect
               word[@sectbuf+$1D8] := startsect>>18
               word[@sectbuf+$1DA] := 8      '8 sectors = 4kByte
               word[@sectbuf+$1DC] := 0
               \sd.writesector(0,@sectbuf)   'write sector 0 with modifications
               dbg.str(string(13,"boot signature written!"))
             showShortcut
           else
             dbg.str(string(13,"no boot file found!",13))

      "4": \sd.readsector(0,@sectbuf)    'read sector 0
           byte[@sectbuf+$1D2] := 0
           word[@sectbuf+$1D6] := 0
           long[@sectbuf+$1D8] := 0
           word[@sectbuf+$1DC] := 0
           \sd.writesector(0,@sectbuf)   'write sector 0 with modifications
           dbg.str(string(13,"boot signature removed!"))
           showShortcut

      "5": \sd.unmount
           dbg.str(string(13,"search.."))
           tm := cnt
           cog := cognew(@entry, 0)
           repeat until (i := long[$7000-4])
           dbg.str(string(".."))
           dbg.dec((cnt-tm)/clkfreq)
           dbg.str(string(" seconds",13))
           if i < 0
             dbg.str(string("booting failed: "))
             dbg.dec(i)
             if i== -9
               dbg.str(string(13,"(no responding SD card)"))
             if i== -10
               dbg.str(string(13,"(no boot signature in sector 0 or file)"))
             if i== -13
               dbg.str(string(13,"(timeout SD access)"))
           else
             dbg.str(string("bootfile at sector: "))
             dbg.dec(i-8)
             dbg.tx(13) 
             byte[$707F] := 0
             dbg.str($7000)
           cogstop(cog)
           dbg.str(string(13,"done (please reload code from IDE)"))
           quit

      "6": showShortcut
      
      "7": err := \sd.popen(string("boot.txt"),"d")
           dbg.str(string(13,"Start..."))
           repeat i from 7 to 19999 step 8
             \sd.readsector(i,@sectbuf)       'test all sectors
             if long[@sectbuf] == long[@sigFile]
               sectbuf[0] := " "
               \sd.writesector(i,@sectbuf)
             \sd.readsector(i+1,@sectbuf)     'test all sectors
             if long[@sectbuf] == long[@sigFile]
               sectbuf[0] := " "
               \sd.writesector(i+1,@sectbuf)
           \sd.readsector(0,@sectbuf)       'read sector 0
           byte[@sectbuf+$1D2] := 0
           word[@sectbuf+$1D6] := 0
           long[@sectbuf+$1D8] := 0
           word[@sectbuf+$1DC] := 0
           \sd.writesector(0,@sectbuf)      'write sector 0 with modifications
           dbg.str(string(".all removed!",13))

PUB directBoot : i
  cog := cognew(@entry, 0)
  repeat until (i := long[$7000-4])
  if i < 0
    dbg.str(string("booting failed: "))
    dbg.dec(i)
    if i== -9
      dbg.str(string(13,"(no responding SD card)"))
    if i== -10
      dbg.str(string(13,"(no boot signature in sector 0 or file)"))
    if i== -13
      dbg.str(string(13,"(timeout SD access)"))
  else
    dbg.str(string("Success! Bootfile at sector: "))
    dbg.dec(i-8)
    dbg.tx(13)
    byte[$707F] := 0
    dbg.str($7000)
   
  repeat                        'stop here

PUB showShortcut : i
  \sd.readsector(0,$6000)
  dbg.str(string(13,"Type: $"))
  dbg.hex(byte[$61D2],2)
  dbg.str(string(13,"boot start sector: "))
  dbg.dec(word[$61D6]+word[$61D8]<<16)
  dbg.str(string(13,"number of sectors: "))
  dbg.dec(word[$61DA]+word[$61DC]<<16)
  dbg.tx(13)

DAT

text  byte "[P2BOOT]",13
      byte "Propeller 2 would load this file from SD-",13
      byte "card. If you see this text the boot file",13
      byte "was found and loaded.",13,0